/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "operator/operator.pmh"

namespace fe
{

void CurveCreateOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	catalog<String>("interpolation")="power";
	catalog<String>("interpolation","label")="Interpolation";
	catalog<String>("interpolation","choice:0")="linear";
	catalog<String>("interpolation","label:0")="Linear";
	catalog<String>("interpolation","choice:1")="pointNormal";
	catalog<String>("interpolation","label:1")="Point Normal";
	catalog<String>("interpolation","choice:2")="cardinal";
	catalog<String>("interpolation","label:2")="Cardinal Spline";
	catalog<String>("interpolation","choice:3")="power";
	catalog<String>("interpolation","label:3")="Power";

	catalog<bool>("bisect")=false;
	catalog<String>("bisect","label")="Bisect";
	catalog<bool>("bisect","joined")=true;

	catalog<bool>("blend")=false;
	catalog<String>("blend","label")="Blend";
	catalog<bool>("blend","joined")=true;

	catalog<Real>("blending")=1.0;
	catalog<String>("blending","label")="Blending";
	catalog<Real>("blending","high")=2.0;
	catalog<Real>("blending","max")=100.0;
	catalog<String>("blending","enabler")="blend";

	catalog<String>("method")="length";
	catalog<String>("method","label")="Method";
	catalog<String>("method","choice:0")="length";
	catalog<String>("method","label:0")="Length";
	catalog<String>("method","choice:1")="count";
	catalog<String>("method","label:1")="Count";

	catalog<Real>("length")=1.0;
	catalog<Real>("length","low")=0.0;
	catalog<Real>("length","high")=1.0;
	catalog<Real>("length","min")=0.0;
	catalog<Real>("length","max")=1e3;
	catalog<String>("length","label")="Segment Length";

	catalog<I32>("segments")=4;
	catalog<I32>("segments","low")=1;
	catalog<I32>("segments","high")=100;
	catalog<I32>("segments","min")=1;
	catalog<I32>("segments","max")=1000000;
	catalog<String>("segments","label")="Segments";

	catalog<Real>("linearity")=0.0;
	catalog<Real>("linearity","max")=1.0;
	catalog<String>("linearity","label")="Linearity";

	catalog<String>("linearityInAttr")="linearityIn";
	catalog<String>("linearityInAttr","label")="Linearity Attr In";
	catalog<bool>("linearityInAttr","joined")=true;
	catalog<String>("linearityInAttr","hint")=
			"Name of real point attribute boosting linearity leading in.";

	catalog<String>("linearityOutAttr")="linearityOut";
	catalog<String>("linearityOutAttr","label")="Out";
	catalog<String>("linearityOutAttr","hint")=
			"Name of real point attribute boosting linearity trailing out.";

	catalog< sp<Component> >("Input Surface");
	catalog<bool>("Input Surface","optional")=true;

	//* no auto-copy
	catalog< sp<Component> >("Output Curve");
	catalog<String>("Output Curve","IO")="output";
}

void CurveCreateOp::handle(Record& a_rSignal)
{
	const String interpolation=catalog<String>("interpolation");
	const BWORD blend=catalog<bool>("blend");
	const Real blending=catalog<Real>("blending");
	const BWORD bisect=catalog<bool>("bisect");
	const String method=catalog<String>("method");
	const Real length=catalog<Real>("length");
	const I32 segments=catalog<I32>("segments");
	const Real linearity=catalog<Real>("linearity");
	const String linearityInAttr=catalog<String>("linearityInAttr");
	const String linearityOutAttr=catalog<String>("linearityOutAttr");

	const BWORD doLinear=(interpolation=="linear");
	const BWORD doCardinal=(interpolation=="cardinal");
	const BWORD doPointNormal=(interpolation=="pointNormal");
	const BWORD doPower=(interpolation=="power");

	const Real nonlinearity=Real(1)-linearity;

	sp<DrawI> spDrawI;
	if(!accessDraw(spDrawI,a_rSignal)) return;

	sp<DrawI> spDrawGuide;
	accessGuide(spDrawGuide,a_rSignal,e_warning);

	sp<SurfaceAccessibleI> spInputAccessible;
	sp<SurfaceAccessorI> spInputPosition;
	sp<SurfaceAccessorI> spInputNormal;
	sp<SurfaceAccessorI> spInputLinearityIn;
	sp<SurfaceAccessorI> spInputLinearityOut;

	access(spInputAccessible,"Input Surface",e_quiet);
	if(spInputAccessible.isValid())
	{
		access(spInputPosition,"Input Surface",
				e_point,e_position,e_quiet);

		if(doPointNormal || doPower)
		{
			access(spInputNormal,spInputAccessible,
					e_point,e_normal,e_warning);
		}

		if(!linearityInAttr.empty())
		{
			access(spInputLinearityIn,spInputAccessible,
					e_point,linearityInAttr,e_warning);
		}

		if(!linearityOutAttr.empty())
		{
			access(spInputLinearityOut,spInputAccessible,
					e_point,linearityOutAttr,e_warning);
		}
	}

	I32 pointCount=0;

	if(spInputPosition.isValid())
	{
		pointCount=spInputPosition->count();
	}

	if(pointCount<2)
	{
		//* no input path

		SpatialVector pointArray[segments+1];

		for(I32 index=0;index<=segments;index++)
		{
			set(pointArray[index],0.001*index*index,index*length);
		}

		spDrawI->drawLines(pointArray,NULL,segments+1,
				DrawI::e_strip,false,NULL);
	}
	else
	{
		Real totalDistance=0.0;
		Real distanceArray[pointCount];

		SpatialVector point0=spInputPosition->spatialVector(0);
		distanceArray[0]=0.0;

		for(I32 pointIndex=1;pointIndex<pointCount;pointIndex++)
		{
			const SpatialVector point1=
					spInputPosition->spatialVector(pointIndex);

			totalDistance+=magnitude(point1-point0);
			distanceArray[pointIndex]=totalDistance;

			point0=point1;
		}

		const I32 trueSegments=(method=="length" && length>0.0)?
				totalDistance/length: segments;

		//* for power modes, precompute cv transforms
		SpatialTransform xformArray[pointCount];
		SpatialTransform middleArray[pointCount];
		if(doPower)
		{
			for(I32 pointIndex=0;pointIndex<pointCount;pointIndex++)
			{
				const I32 cv1=pointIndex;
				const I32 cv0=fe::maximum(cv1-1,0);
				const I32 cv00=fe::maximum(cv1-2,0);
				const I32 cv2=fe::minimum(cv1+1,pointCount-1);
				const I32 cv3=fe::minimum(cv1+2,pointCount-1);

				const SpatialVector p00=spInputPosition->spatialVector(cv00);
				const SpatialVector p0=spInputPosition->spatialVector(cv0);
				const SpatialVector p1=spInputPosition->spatialVector(cv1);
				const SpatialVector p2=spInputPosition->spatialVector(cv2);
				const SpatialVector p3=spInputPosition->spatialVector(cv3);

				SpatialVector n0(0.0,1.0,0.0);
				SpatialVector n1(0.0,1.0,0.0);
				SpatialVector n2(0.0,1.0,0.0);

				if(spInputNormal.isValid())
				{
					n0=spInputNormal->spatialVector(cv0);
					n1=spInputNormal->spatialVector(cv1);
					n2=spInputNormal->spatialVector(cv2);
				}

				//* vertex tangent is parallel to connection of neighbors
//				const SpatialVector t1=unitSafe(p2-p0);

				//* vertex tangent is average of incoming and outgoing segments
				const SpatialVector t1=
						unitSafe(unitSafe(p2-p1)+unitSafe(p1-p0));

				//* vertex tangent uses sqrt between neighbors
//				const SpatialVector t0=
//						(cv1>cv00)? unitSafe(p1-p00): unitSafe(p2-p00);
//				const SpatialVector t2=
//						(cv3>cv1)? unitSafe(p3-p1): unitSafe(p3-p0);

//				SpatialTransform xform0;
//				makeFrameTangentX(xform0,p0,t0,n0);
//
//				SpatialTransform xform2;
//				makeFrameTangentX(xform2,p2,t2,n2);
//
//				SpatialTransform xform1;
//				halfTransform(xform1,xform0,xform2);
//
//				SpatialTransform xform1plus;
//				interpolateTransform(xform1plus,xform0,xform2,0.51);
//
//				const SpatialVector t1=unitSafe(
//						xform1plus.translation()-xform1.translation());

				makeFrameTangentX(xformArray[pointIndex],p1,t1,n1);
			}

			if(bisect)
			{
				for(I32 pointIndex=0;pointIndex<pointCount-1;pointIndex++)
				{
					const I32 cv1=pointIndex;
					const I32 cv2=fe::minimum(cv1+1,pointCount-1);

					const SpatialVector p1=spInputPosition->spatialVector(cv1);
					const SpatialVector p2=spInputPosition->spatialVector(cv2);

					SpatialVector n1(0.0,1.0,0.0);
					SpatialVector n2(0.0,1.0,0.0);

					if(spInputNormal.isValid())
					{
						n1=spInputNormal->spatialVector(cv1);
						n2=spInputNormal->spatialVector(cv2);
					}

					const SpatialTransform xform1=xformArray[pointIndex];
					const SpatialTransform xform2=xformArray[pointIndex+1];

					//* halfway point is for position only
					SpatialTransform halfway;
					halfTransform(halfway,xform1,xform2);

					const SpatialVector ptMid=halfway.translation();
					const SpatialVector nMid=unitSafe(n1+n2);

#if FALSE
					const SpatialVector tMid=unitSafe(p2-p1);
#else
					const SpatialVector tMidPre=unitSafe(p2-p1);

					SpatialTransform middlePlus;
					makeFrameTangentX(middlePlus,ptMid,tMidPre,nMid);

					SpatialTransform halfwayPlus;
					interpolateTransform(halfwayPlus,middlePlus,xform2,0.01);

					const SpatialVector pMidPlus=halfwayPlus.translation();

					const SpatialVector tMid=unitSafe(pMidPlus-ptMid);
#endif

					makeFrameTangentX(middleArray[pointIndex],ptMid,tMid,nMid);
				}
			}
			middleArray[pointCount-1]=middleArray[pointCount-2];
		}

		SpatialVector pointArray[trueSegments+1];

		I32 cv2=0;

		for(I32 index=0;index<=trueSegments;index++)
		{
			const Real along=index/Real(trueSegments);

			const Real distance=along*totalDistance;

			while(distanceArray[cv2]<=distance && cv2<pointCount-1)
			{
				cv2++;
			}

			const I32 cv1=fe::maximum(cv2-1,0);

			const Real t=(distance-distanceArray[cv1])/
					(distanceArray[cv2]-distanceArray[cv1]);
			const Real t1=Real(1)-t;

			const SpatialVector p1=spInputPosition->spatialVector(cv1);
			const SpatialVector p2=spInputPosition->spatialVector(cv2);

			const SpatialVector ptLinear=p1*t1+p2*t;

			if(doLinear)
			{
				pointArray[index]=ptLinear;
				continue;
			}

			Real nonlinearityOut1=1.0;
			Real nonlinearityIn2=1.0;
			if(spInputLinearityOut.isValid())
			{
				nonlinearityOut1=Real(1)-spInputLinearityOut->real(cv1);
			}
			if(spInputLinearityIn.isValid())
			{
				nonlinearityIn2=Real(1)-spInputLinearityIn->real(cv2);
			}

			const Real oneNonlinearity=
					nonlinearity*(nonlinearityOut1*t1+nonlinearityIn2*t);
			const Real oneLinearity=Real(1)-oneNonlinearity;

			if(doPointNormal)
			{
				SpatialVector n1(0.0,1.0,0.0);
				SpatialVector n2(0.0,1.0,0.0);

				if(spInputNormal.isValid())
				{
					n1=spInputNormal->spatialVector(cv1);
					n2=spInputNormal->spatialVector(cv2);
				}

				TrianglePN<Real> trianglePN;
				trianglePN.configure(p1,p2,p2,n1,n2,n2);

				const SpatialBary bary(t1,t,Real(0));
				SpatialVector point;
				SpatialVector norm;
				trianglePN.solve(bary,point,norm);

				pointArray[index]=oneLinearity*ptLinear+oneNonlinearity*point;
				continue;
			}

			const I32 cv0=fe::maximum(cv2-2,0);
			const I32 cv3=fe::minimum(cv2+1,pointCount-1);

			if(doCardinal)
			{
				const SpatialVector p0=spInputPosition->spatialVector(cv0);
				const SpatialVector p3=spInputPosition->spatialVector(cv3);

				pointArray[index]=oneLinearity*ptLinear+oneNonlinearity*
						Spline<SpatialVector,Real>::Cardinal2D(t,p0,p1,p2,p3);
				continue;
			}

			const SpatialTransform xform0=xformArray[cv0];
			const SpatialTransform xform1=xformArray[cv1];
			const SpatialTransform xform2=xformArray[cv2];
			const SpatialTransform xform3=xformArray[cv3];

			SpatialTransform between;
			SpatialTransform middle;

			SpatialTransform xformA=xform0;
			SpatialTransform xformB=xform1;
			SpatialTransform xformC=xform2;
			SpatialTransform xformD=xform3;
			Real tt=t;

			if(bisect)
			{
				if(t<0.5)
				{
					//* first half
					xformA=middleArray[cv0];

//*					xformB=xform1;	//* NOTE already is

					middle=middleArray[cv1];
					xformC=middle;

					xformD=xform2;

					tt=t*Real(2);
				}
				else
				{
					//* second half
					xformA=xform1;

					middle=middleArray[cv1];
					xformB=middle;

//*					xformC=xform2;	//* NOTE already is

					xformD=middleArray[cv2];

					tt=t*Real(2)-Real(1);
				}
			}

			if(blend)
			{
				const Real distBC=magnitude(
						xformC.translation()-xformB.translation());

				const Real leftT=pow(tt,blending);
				const Real rightT=Real(1)-pow(Real(1)-tt,blending);

#if TRUE
				//* NOTE keep curving

				const Real distAB=magnitude(
						xformB.translation()-xformA.translation());
				const Real distCD=magnitude(
						xformD.translation()-xformC.translation());

				//* accounting for non-uniform CV spacing
				const Real scaleAB=fe::minimum(1.0,
						(distAB>0.0)? distBC/distAB: 1.0);
				const Real scaleCD=fe::minimum(1.0,
						(distCD>0.0)? distBC/distCD: 1.0);

				SpatialTransform betweenAB;
				interpolateTransform(betweenAB,xformA,xformB,
						Real(1)+scaleAB*tt);

				SpatialTransform betweenCD;
				interpolateTransform(betweenCD,xformD,xformC,
						Real(1)+scaleCD*(Real(1)-tt));
#else
				//* NOTE extrapolate in a line

				SpatialTransform slightAB;
				interpolateTransform(slightAB,xformA,xformB,1.01);

				SpatialTransform slightCD;
				interpolateTransform(slightCD,xformD,xformC,1.01);

				const SpatialVector dirAB=unitSafe(
						slightAB.translation()-xformB.translation());
				const SpatialVector dirCD=unitSafe(
						slightAB.translation()-xformB.translation());

				SpatialTransform betweenAB=xformB;
				SpatialTransform betweenCD=xformC;

				betweenAB.translation()+=tt*distBC*dirAB;
				betweenCD.translation()+=(Real(1)-tt)*distBC*dirCD;
#endif

				SpatialTransform betweenBC;
				interpolateTransform(betweenBC,xformB,xformC,tt);

				SpatialTransform betweenLeft;
				interpolateTransform(betweenLeft,betweenAB,betweenBC,leftT);

				SpatialTransform betweenRight;
				interpolateTransform(betweenRight,betweenBC,betweenCD,rightT);

				halfTransform(between,betweenLeft,betweenRight);
			}
			else
			{
				interpolateTransform(between,xformB,xformC,tt);
			}

			pointArray[index]=
					oneLinearity*ptLinear+oneNonlinearity*between.translation();

			if(spDrawGuide.isValid())
			{
				const Color lightred(1.0,0.3,0.3,1.0);
				const Color lightgreen(0.3,1.0,0.3,1.0);
				const Color lightblue(0.3,0.3,1.0,1.0);

				SpatialVector line[2];

				const U32 passes=bisect? 4: 3;

				for(U32 pass=0;pass<passes;pass++)
				{
					const SpatialTransform step=
							pass==0? xform1:
							(pass==1? xform2:
							(pass==2? between: middle));

					const Real guidelength=10.0;
					const Real scalar=guidelength*
							(pass<2? 1.0: (pass==2? 0.3: 0.5));

					line[0]=step.translation();

					for(U32 n=0;n<3;n++)
					{
						const Color color=
								n==0? lightred: (n==1? lightgreen: lightblue);

						line[1]=line[0]+scalar*step.column(n);

						spDrawGuide->drawLines(line,NULL,2,
								DrawI::e_strip,false,&color);
					}
				}
			}
		}

		spDrawI->drawLines(pointArray,NULL,trueSegments+1,
				DrawI::e_strip,false,NULL);
	}
}

//* static
void CurveCreateOp::halfTransform(SpatialTransform& a_result,
	const SpatialTransform& a_matrix1,const SpatialTransform& a_matrix2)
{
	MatrixSqrt<SpatialTransform> matrixSqrt;

	SpatialTransform inv1;
	invert(inv1,a_matrix1);

	const SpatialTransform delta12=inv1*a_matrix2;

	SpatialTransform partial12;
	matrixSqrt.solve(partial12,delta12);

	a_result=a_matrix1*partial12;
}

//* static
void CurveCreateOp::interpolateTransform(SpatialTransform& a_result,
	const SpatialTransform& a_matrix1,const SpatialTransform& a_matrix2,
	Real a_fraction)
{
	MatrixPower<SpatialTransform> matrixPower;

	SpatialTransform inv1;
	invert(inv1,a_matrix1);

	const SpatialTransform delta12=inv1*a_matrix2;

	SpatialTransform partial12;
	matrixPower.solve(partial12,delta12,a_fraction);

	a_result=a_matrix1*partial12;
}

} // namespace
